////////////////////////////////////////////////////////////////////////////////////////
// File: scripts/custom_explosion.js
// Usage: custom explosion logic with hardcoded resistance values
////////////////////////////////////////////////////////////////////////////////////////

import { world } from "@minecraft/server";

////////////////////////////////////////////////////////////////////////////////////////

export function createCustomExplosion(event) {
  const origin = event.sourceEntity.location;
  const [power, drop, fire] = event.message.split(",").map(value => {
    value = value.replace(/"/g, '').replace(/\s+/g, '');
    if (value === "true") return true;
    if (value === "false") return false;
    return isNaN(value) ? value : Number(value);
  });
  if (power <= 0) return;

  // cache dimension
  const dimension = event.sourceEntity.dimension

  // Range for filling air at start
  const fillRanges = [
    { min: 5, max: 8, offset: 0 },
    { min: 9, max: 15, offset: 1 },
    { min: 16, max: 20, offset: 2 },
    { min: 21, max: 32, offset: 3 },
    { min: 33, max: Infinity, offset: 4 }
  ];
  for (const { min, max, offset } of fillRanges) {
    if (power >= min && power <= max) { // fill air of correct size at start to reduce later calculations
      dimension.runCommand(`/fill ${origin.x - offset} ${origin.y - offset + 1} ${origin.z - offset} ${origin.x + offset} ${origin.y + offset + 1} ${origin.z + offset} air replace`);
      break;
    }
  }

  // Damage handling
  if (power >= 3) {
    const maxRadius = Math.round(Math.min(Math.max(power * 0.8 + 1, 2), 32));
    const maxDmg = Math.round(Math.max(6 * Math.sqrt(power), 10));

    // total num of explosion steps
    const steps = Math.round(power / 3) + 1;

    for (let i = 1; i <= steps; i++) {
      let r = Math.round(maxRadius * (i / steps));
      let dmg = Math.round(maxDmg * ((steps - i + 1) / steps));

      event.sourceEntity.runCommand(`/execute positioned ${origin.x} ${origin.y} ${origin.z} run damage @e[r=${r},type=!armor_stand,type=!item,family=!utility] ${dmg} entity_explosion entity @s`);
      event.sourceEntity.runCommand(`/execute positioned ${origin.x} ${origin.y} ${origin.z} as @e[r=${2 * r},type=!armor_stand,type=!item,family=!utility] run scriptevent sbs_mib:explosion_knockback "${Math.round(origin.x)}, ${Math.round(origin.y)}, ${Math.round(origin.z)}, ${2 * r}, ${dmg}"`);
    }
  }


  // directions and distance
  const directions = generateDirections(1352);
  const maxDistanceSquared = power * power;

  directions.forEach(direction => {
    let currentPos = { ...origin };
    let remainingPower = 2 * power / 0.225
    let traveledDistanceSquared = 0;

    while (remainingPower > 0 && traveledDistanceSquared < maxDistanceSquared) {
      // Calculate next position
      currentPos.x += direction.x;
      currentPos.y += direction.y;
      currentPos.z += direction.z;

      const block = dimension.getBlock(currentPos);  // Get the block once per iteration
      if (!block) break;  // Stop ray if out of bounds or in an unloaded chunk

      const resistance = getBlockResistance(block);
      if (resistance === null) break;
      remainingPower -= Math.pow(resistance, 0.55) * 10
      if (remainingPower <= 0) break;  // Stop if power is depleted

      try {
        if (block && block.typeId !== 'minecraft:air') {
          const shouldDrop = drop && Math.random() < (fire ? (2 / power) : (1 / power));
          if (fire && Math.random() < 0.35) {
            block.setType('minecraft:fire');
          } else if (shouldDrop) {
            block.dimension.runCommand(`/setblock ${block.location.x} ${block.location.y} ${block.location.z} air destroy`);
          } else {
            block.setType('minecraft:air');
          }
        }
      } catch (error) {
        break;
      }

      // Update traveled distance
      traveledDistanceSquared = (currentPos.x - origin.x) ** 2 +
        (currentPos.y - origin.y) ** 2 +
        (currentPos.z - origin.z) ** 2;
    }
  });
}

////////////////////////////////////////////////////////////////////////////////////////

export function explosionKnockback(event) {
  const entity = event.sourceEntity;

  const explosion = event.message.split(",").map(value => {
    const trimmedValue = value.trim().replace(/"/g, '');
    return !isNaN(trimmedValue) ? Number(trimmedValue) : trimmedValue;
  }).reduce((acc, val, idx) => {
    acc[['x', 'y', 'z', 'r', 'dmg'][idx]] = val;
    return acc;
  }, {});

  if (explosion.r <= 0) return;

  const direction = {
    x: entity.location.x - explosion.x,
    y: entity.location.y - explosion.y,
    z: entity.location.z - explosion.z,
  };

  const distance = Math.sqrt(direction.x ** 2 + direction.y ** 2 + direction.z ** 2);

  if (distance <= explosion.r) {
    const distanceRatio = distance / explosion.r; // Ratio from 0 (center) to 1 (edge)
    const scaledRatio = 1 - Math.pow(distanceRatio, 0.3); // More sensitive scaling for closer entities

    const horizontalStrength = scaledRatio * 1.5 * explosion.dmg; // Reduced max horizontal strength
    const verticalStrength = scaledRatio * (direction.y > 0 ? 0.5 * explosion.dmg : 0); // Lower scaling for vertical

    const normalizedDir = {
      x: direction.x / distance,
      z: direction.z / distance,
    };

    entity.applyKnockback(normalizedDir.x, normalizedDir.z, horizontalStrength, verticalStrength);
  }
}



////////////////////////////////////////////////////////////////////////////////////////

/**
 * Generates directions for rays based on spherical coordinates.
 * @param {number} count - The number of directions to generate.
 * @returns {Vector3[]} Array of directional vectors.
 */
function generateDirections(count) {
  return Array.from({ length: count }, () => {
    const theta = Math.acos(2 * Math.random() - 1);
    const phi = 2 * Math.PI * Math.random();
    return {
      x: Math.sin(theta) * Math.cos(phi),
      y: Math.sin(theta) * Math.sin(phi),
      z: Math.cos(theta),
    };
  });
}

////////////////////////////////////////////////////////////////////////////////////////

const resistanceTable = Object.freeze({
  "minecraft:stone": 6,
  "minecraft:dirt": 2,
  "minecraft:glass": 0.3,
  "minecraft:wood": 3,
  "minecraft:oak_planks": 3,
  "minecraft:sand": 2,
  "minecraft:gravel": 1.5,
  "minecraft:cobblestone": 12,
  "minecraft:obsidian": 1200,
  "minecraft:bedrock": Infinity,
  "minecraft:iron_block": 30,
  "minecraft:diamond_block": 50,
  "minecraft:gold_block": 15,
  "minecraft:netherite_block": 1050,
  "minecraft:clay": 1.5,
  "minecraft:soul_sand": 3,
  "minecraft:soul_soil": 3,
  "minecraft:quartz_block": 15,
  "minecraft:brick_block": 30,
  "minecraft:mossy_cobblestone": 10,
  "minecraft:stone_bricks": 30,
  "minecraft:chiseled_stone_bricks": 30,
  "minecraft:cracked_stone_bricks": 30,
  "minecraft:purpur_block": 10,
  "minecraft:prismarine": 15,
  "minecraft:dark_prismarine": 15,
  "minecraft:sea_lantern": 15,
  "minecraft:nether_bricks": 30,
  "minecraft:red_nether_brick": 30,
  "minecraft:basalt": 10,
  "minecraft:blackstone": 15,
  "minecraft:polished_blackstone": 15,
  "minecraft:polished_blackstone_bricks": 15,
  "minecraft:ancient_debris": 1050,
  "minecraft:crying_obsidian": 600,
  "minecraft:glowstone": 3,
  "minecraft:shroomlight": 3,
  "minecraft:end_stone": 15,
  "minecraft:end_stone_bricks": 15,
  "minecraft:purpur_pillar": 10,
  "minecraft:purpur_slab": 10,
  "minecraft:slime_block": 0.3,
  "minecraft:hay_block": 0.3,
  "minecraft:carved_pumpkin": 3,
  "minecraft:jack_o_lantern": 3,
  "minecraft:melon": 0.3,
  "minecraft:sponge": 3,
  "minecraft:wet_sponge": 3,
  "minecraft:mycelium": 2,
  "minecraft:podzol": 2,
  "minecraft:flower_pot": 0.3,
  "minecraft:grass_block": 2,
  "minecraft:stone_slab": 6,
  "minecraft:stone_stairs": 6,
  "minecraft:wooden_slab": 3,
  "minecraft:wooden_stairs": 3,
  "minecraft:smooth_stone_slab": 6,
  "minecraft:brick_slab": 30,
  "minecraft:stone_button": 6,
  "minecraft:wooden_button": 3,
  "minecraft:anvil": 100,
  "minecraft:chiseled_quartz_block": 15,
  "minecraft:quartz_slab": 15,
  "minecraft:quartz_stairs": 15,
  "minecraft:quartz_pillar": 15,
  "minecraft:chain": 15,
  "minecraft:lantern": 15,
  "minecraft:iron_bars": 15,
  "minecraft:glass_pane": 0.3,
  "minecraft:carpet": 0.1,
  "minecraft:bed": 10,
  "minecraft:trapdoor": 3,
  "minecraft:wooden_trapdoor": 3,
  "minecraft:iron_trapdoor": 15,
  "minecraft:oak_door": 3,
  "minecraft:iron_door": 15,
  "minecraft:ladder": 3,
  "minecraft:fence": 3,
  "minecraft:nether_fence": 30,
  "minecraft:spruce_fence": 3,
  "minecraft:birch_fence": 3,
  "minecraft:dark_oak_fence": 3,
  "minecraft:jungle_fence": 3,
  "minecraft:acacia_fence": 3,
  "minecraft:brick_wall": 30,
  "minecraft:stone_wall": 30,
  "minecraft:mossy_stone_wall": 30,
  "minecraft:iron_wall": 15,
  "minecraft:snow_block": 0.1,
  "minecraft:light_weighted_pressure_plate": 15,
  "minecraft:heavy_weighted_pressure_plate": 15,
  "minecraft:pressure_plate": 3,
  "minecraft:stone_pressure_plate": 6,
  "minecraft:wooden_pressure_plate": 3,
  "minecraft:emerald_block": 50,
  "minecraft:lapis_block": 15,
  "minecraft:dispenser": 3,
  "minecraft:dropper": 3,
  "minecraft:chest": 3,
  "minecraft:trapped_chest": 3,
  "minecraft:barrel": 3,
  "minecraft:shulker_box": 3,
  "minecraft:end_chest": 3,
  "minecraft:hopper": 3,
  "minecraft:loom": 3,
  "minecraft:smithing_table": 3,
  "minecraft:blast_furnace": 3,
  "minecraft:furnace": 3,
  "minecraft:smoker": 3,
  "minecraft:cartography_table": 3,
  "minecraft:stonecutter": 3,
  "minecraft:composter": 3,
  "minecraft:enchanted_book": 0,
  "minecraft:spawner": 3,
  "minecraft:mob_spawner": 3,
  "minecraft:end_portal_frame": 1000,
  "minecraft:end_portal": 1000,
  "minecraft:respawn_anchor": 1000,
  "minecraft:command_block": 1000,
  "minecraft:repeating_command_block": 1000,
  "minecraft:chain_command_block": 1000,
  "minecraft:structure_block": 1000,
  "minecraft:barrier": 1000,
  "minecraft:world_border": 1000,
  "minecraft:light_block": 1000,
});

function getBlockResistance(block, defaultResistance = 1) {
  return resistanceTable[block.typeId] || defaultResistance;
}

////////////////////////////////////////////////////////////////////////////////////////